#!/usr/bin/perl

# order.pl

$version = "1.0.9";
$ftpservername = "ftp.phonon.net";

# This script was written by Riad Wahby.
# It is hereby entered into the public
# domain under the provisions of the GNU
# public license.  It may be freely 
# distributed as long as this header remains
# intact.  Direct feature requests to
# rsw@phonon.net

# This is a generic drop-in script for use
# with Total Merchant Manager Virtual Terminal
# v.3.x systems.

require 5.001;
use LWP::UserAgent;
use HTTP::Request;
use HTTP::Response;
use Net::SMTP;

sub admin_error; # error response for invalid
                 # setup.  Prints an HTML reference
                 # for setup and use of this script

sub user_error; # error response for incomplete
                # input.  This is triggered by
                # required fields missing or a
                # total of zero items

sub send_mail;  # sends an e-mail message

sub http_dump;  # returns response from URL

sub get_colors; # gets the color and image options
                # from the post.

sub read_post;  # reads information from POST

sub check_post; # checks information for 
                # fields from @requiredfields

sub check_opts; # checks information for
                # required admin options

sub check_price; # checks the total order price

sub check_qty;  # checks the total order quantity

sub check_ship; # checks the total shipping price

sub determine_shipprice; # from the input of shipping
                # and international shipping, it determines
                # what price the customer should be charged

sub print_taxes; # checks whether the state
                 # is one which requires tax
                 # and adds the tax if so

sub print_items; # prints the items in HTML format for TMM

sub print_bill; # prints the billing address

sub print_ship; # prints an input for an alternate shipping address

sub print_header; # prints the HTTP header and the first few HTML tags

sub print_form_setup; # prints the HTML tags for the form, setting all
                      # necessary TMM options

sub returning_from_mailto; # sends the mail from the form, and goes to mail_result_url

sub returning_from_TMM; # gives the user message based on what the credit card 

sub mail_order_dispatch; # dispatch for orders to be sent by e-mail

sub TMM_order_dispatch; # dispatch for orders to be sent using TMM

################# REQUIREMENTS ##################

# This script requires Net::SMTP, which is not
# part of the standard library modules.  It
# is available from ftp.cpan.org in path
# /pub/CPAN/modules/by-module/Net/libnet-1.0606.tar.gz

# HTTP and LWP, also required, can be found in 
# libwww-perl, which is available at
# http://www.sn.no/libwww-perl/ or
# ftp.cpan.org/pub/CPAN/modules/by-module/HTTP/

# If you need some help, reference the script
# with absolutely no options.  This will print
# an HTML command reference.  If you reference
# it with administrative options, it will
# assume bad user input and attempt to create
# a user "error" page.

################# MAIN PROGRAM #################

read_post();

if ($FORM{'mailto'})
{
    returning_from_mailto();
}
elsif ($FORM{'validated'})
{
    returning_from_TMM();
}

unless (check_opts())
{
    admin_error();
}

get_colors();

unless (check_post())
{
    user_error("You did not fill in all required fields.");
}

unless (check_qty())
{
    user_error("You did not make a product selection.");
}

if (check_price())
{
    TMM_order_dispatch();
}
else
{
    mail_order_dispatch();
}

################# SUBPROGRAMS #################

sub admin_error
{
#
# Takes no input
#
# Always kills the script execution
#
    print <<END;
Content-type: text/html

<html><head><title>TMMOrder $version</title></head>
<body bgcolor=#ffffff>
<h1>TMMOrder $version</h1>

<p>You must provide several options to this script as hidden fields on
the submission form.  This will allow the script to operate in a way
most compatible with your site.  If you are unsure why you are
receiving this message, it is most probably because you did not
include all required arguments.  Please check your setup.

<h3>General Use</h3>
<p>To use this script, all that is necessary is that arguments be POSTed to it using the HTML &lt;form&gt; tag.  For instance,
<pre>
&lt;form method=POST action=http://foo.com/cgi-bin/tmmorder.pl&gt;
</pre>
<p>Following this, all options must be POSTed as hidden or input fields.

<h3>Required Administrative Fields</h3>
<p>The following fields are required.  If they are not present, this error
message will be displayed.
<ul>
<li><pre>result_url</pre> - the result URL when credit card validation is successful -- see note below<br><br>
<li><pre>result_url_unverified</pre> - the result URL when credit card validation is unsuccessful -- see note below<br><br>
<li><pre>vt_url</pre> - the URL of the TMM Virtual Terminal which will be used to validate credit cards<br><br>
<li><pre>mail_company</pre> - the name of the company that is receiving payment (your company)<br><br>
<li><pre>mail_address</pre> - the street address of the above<br><br>
<li><pre>mail_city</pre> - the city of the above<br><br>
<li><pre>mail_state</pre> - the state of the above<br><br>
<li><pre>mail_zip</pre> - the zip code of the above<br><br>
<li><pre>mail_recipient</pre> - the e-mail address of the person who should receive the orders<br><br>
<li><pre>script_url</pre> - the URL of this script<br><br>
<li><pre>local_country</pre> - the name of the country in which the receiving company (your company) is located<br><br>
<li><pre>mail_result_url</pre> - the URL of the result page when the output of the script is a mail to you
without TMM interaction (when the total order is $0.00 but there is one or more items--i.e. for free items)<br><br>
<li><pre>mailserver</pre> - the mailserver to use when the order is e-mailed instead of going through TMM.  This
happens when there are items being ordered, but the total cost is zero (i.e. request of a free trial, etc)<br><br>
</ul>

<h3>Optional Administrative Fields</h3>
<p>The following fields allow finer control over the behavior of the script.
<ul>
<li><pre>bgcolor</pre> - the background color of all pages generated.  Default is FFFFFF. It must be in RGB 
hex triplet format, as with the bgcolor attribute of the <body> tag, but without the leading hash mark.
Thus, for black, use
<pre>&lt;input type=hidden name=bgcolor value=000000&gt;</pre><br><br>
<li><pre>textcolor</pre> - the text color of all pages generated.  Default is 000000<br><br>
<li><pre>bg_image</pre> - the background image for all pages generated Default is same as bgcolor.<br><br>
<li><pre>no_print_validated</pre> - Normally, when a user is validated, this script prints a message saying that
the person was validated and gives the validation code, then prints the contents of result_url after this.  If
you would rather that TMM returns to a different page (a different script, for instance), set this variable to "true".
Otherwise do not set this variable<br><br>
<li><pre>head_image</pre> - the URL of the image that should be printed at the top of all output pages<br><br>
<li><pre>table_bgcolor</pre> - the background color of the table containing the items that the person
is ordering<br><br>
<li><pre>cc_recipient</pre> - recipients of copies of the e-mails generated by this script (but not by TMM) in a 
comma-delimited list, i.e. "foo\@bar.com,user\@domain.com"<br><br>
</ul>

<h3>Required User Input Fields</h3>
<p>The following fields must be entered by the user or the script will generate an error.
<ul>
<li><pre>snd_name</pre> - the name of the user<br><br>
<li><pre>snd_addr1</pre> - the address of the user<br><br>
<li><pre>snd_city</pre> - the city of the user<br><br>
<li><pre>snd_state</pre> - the state of the user (2-letter postal abbreviations only)<br><br>
<li><pre>snd_zip</pre> - the zip code of the user<br><br>
<li><pre>snd_country</pre> - the country of the user<br><br>
<li><pre>snd_phone</pre> - the phone number of the user<br><br>
<li><pre>snd_mail</pre> - the e-mail address of the user<br><br>
</ul>
<p> You can also make other fields required.  In order to do this, make a hidden field called [name]_required
where [name] is the name of the field.  For instance, to require the user to enter the last_name field, use
    <pre> &lt;input type="hidden" name="last_name_required" value="true"&gt; </pre>

<h3>Items</h3>
<p>The following fields are required for each item:
<ul>
<li><pre>itemXnum</pre> - the item number (not quantity) for item X.  Replace X with sequential numbers starting
with 1<br><br>
<li><pre>itemXprice</pre> - the price each for item X<br><br>
<li><pre>itemXqty</pre> - the quantity of item X being purchased<br><br>
<li><pre>itemXdesc</pre> - a string description of item X<br><br>
<li><pre>itemXship, itemXiship</pre> - see description below in <strong>Shipping</strong><br><br>
</ul>

<h3>Tax</h3>
<p>The following is how to specify tax rates
<ul>
<li><pre>[STATE]_tax</pre> - to specify the tax rate for a certain state, use this form, where [STATE] is the two-letter
postal abbreviation for the state.  The value of the input field should be the decimal tax rate.  For instance, to
specify that purchasers in Iowa pay 5% tax, use the following: 
<pre> &lt;input type="hidden" name="IA_tax" value=".05"&gt; </pre>
You may specify as many tax rates as you like.  Users who are from unlisted states pay no tax.
</ul>

<h3>Shipping</h3>
<p>If you want to give the option to use an alternate shipping address, simply add a checkbox named "diffship" as
in this example:

<pre> Check here if you want this order shipped to a different address:
    &lt;input type="checkbox" name="diffship"&gt; </pre>

<p>Each option for specifiying shipping prices should be followed by a number (the X in each option below).
This number is the shipping method.  ship_method should be used to choose which rate is desirable.  If there
is only one shipping rate, provide only option 1 and ship_method will be automatically set to 1 by the script.
<ul>
<li><pre>ship_method</pre> - a radio button which determines which shipping price will be used.  If omitted,
method 1 is used<br><br>
<li><pre>ship_method_X_name</pre> - a hidden field which should be included for each shipping method which
provides the name of the shipping method (i.e. "2nd day", "first class", etc). If you are specifying shipping 
price by item, you must specify the price for each shipping method on every item<br><br>
</ul>
<p>Choose one of the following methods for determining shipping rate.  The methods are listed in order of
precedence; if two methods are used, the one of higher precedence will be the one that actually determines
the rate for shipping.  
<ol>
<li><ul>
<li><pre>global_shipX</pre> - The total shipping price.  Only enter this field if you want the shipping rate
to remain the same regardless of the order<br><br>
<li><pre>global_ishipX</pre> - The total international shipping price.  Enter this along with the above to
indicate a difference in rate to countries other than the one specified in local_country.  Leave unused if the
rate is the same to all countries<br><br>
</ul>
<li><ul>
<li><pre>global_ship_eachX</pre> - The shipping rate for each item.  Only enter this field if each item ordered has
the same shipping rate.  global_ship_each * totalquantity gives the shipping cost<br><br>
<li><pre>global_iship_eachX</pre> - The international shipping rate for each item.  Enter this field along with
the above to indicate a difference in rate to countries other than the one specified in local_country.  Leave
unused if the rate is the same to all countries<br><br>
</ul>
<li><ul>
<li><pre>itemXshipX</pre> - the shipping rate for each item X.  Enter this field if you want each item to have
a different shipping rate.  If you want to use this for one item, you must use it for every item.  Total shipping
is determined by itemXship * itemXqty + itemYship * itemYqty etc.<br><br>
<li><pre>itemXishipX</pre> - the shipping rate for each item X.  Enter this field along with the above to indicate
a difference in rate to countries other than the one specified in local_country.  Leave unused if the rate is
the same to all countries<br><br>
</ul></ol>

<h3>Returning to CGI Scripts</h3>
<p>It is possible to make the Total Merchant Manager Virtual Terminal return directly to result_url instead of
printing the validation information and then dumping the appropriate HTML.  This is useful if you want to return
to a CGI script, such as one that will add a user to a database.  If it is necessary to pass variables based on
the user input, the following methods of setting variables to pass through the TMM Virtual Terminal are available:
<ul>
<li><pre>passthruX</pre> - to always pass a variable through, use this tag.  X is any number 1 - 9<br><br>
<li><pre>passvarX</pre> - to pass the value of the variable specified as the value of this option through the script, 
use this variant of the above.  Any number used as X here may not be used above.  Valid X is 1 - 9, excluding 
values used above.
</ul>
<p>These will be passed through as POST variables USER1 - USER9.  passthruX corresponds to USERX.
<p>Total Merchant Manager returns the following variables:
<ul>
<li><pre>VALIDATED = Y | N</pre>
<li><pre>APP_CODE = approval code</pre>
</ul>
<p>These indicate status of validation (Y is validated, N is not), and the approval code if the card was validated.
TMM does not return the credit card number unless there is a secure URL whose SSL certificate matches the Merchant
Name.  Contact TMM for more information.

<h3>Note on result_url and result_url_unverified</h3>

<p>The HTML page pointed to by result_url or result_url_unverified should not be a full HTML page, 
but rather the latter half of one.  The output of the script will print out the beginning tags, 
and state whether the credit card was validated or not, then dump the HTML code from either result_url
or result_url_unverified as appropriate.
<p>The above is only true if no_print_validated is unset.  If no_print_validated is set, result_url must be either
a full HTML page or a CGI script, and result_url_unverified is ignored.  
<p>See the section <strong>Returning to CGI Scripts</strong> above for more information.
<p><hr>

<p>This script was written by <a href="mailto:rsw\@phonon.net">Riad Wahby (rsw\@phonon.net)</a>.  I provide no warranties
of any kind.  Any damage done by this script is not my responsibility.  If you have any questions or suggestions
for improving this script, please mail them to me.</p>
<p>TMMOrder version $version
<p>The newest version is always available at 
<a href="ftp://$ftpservername/pub/tmmorder">ftp://$ftpservername/pub/tmmorder</a>.
</body>
</html>
END
    exit;
}

sub user_error
{
#
# Takes one argument, the error message to print.
#
# Always kills script execution
#   
    print_header("Error: Form Not Complete");
    print <<END;
<p>$_[0]  Please use the back button in your browser to return to the form.
</body>
</html>
END
    exit;
}

sub send_mail
{
#
# Takes arguments: 
# $message (message, as a string)
# $subject (subject string)
# $to (To: string)
# $from (From: string)
# $cc (CC: string, optional, comma-delimited)
#
# Upon success, it returns 1.  Otherwise, it
# returns 0
#

    my $message = $_[0];
    my $subject = $_[1];
    my $to = $_[2];
    my $from = ($_[3]);
    my @cc = split(',', $_[4]);
    $to =~ s/\s//g;
    if (my $smtp = Net::SMTP->new($FORM{'mailserver'}))
    {
	$smtp->mail($from);
	$smtp->to($to);
	$smtp->to($from);
	if (@cc)
	{
	    foreach $cc (@cc)
	    {
		$cc =~ s/\s//g;
		$smtp->to($cc);
	    }
	}
	$smtp->data();
	$smtp->datasend("To: $to\n");
	$smtp->datasend("CC: $from\n");
	if (@cc)
	{
	    foreach $cc (@cc)
	    {
		$smtp->datasend("CC: $cc\n");
	    }
	}
	$smtp->datasend("From: $from\n");
	$smtp->datasend("Subject: $subject\n");
	$smtp->datasend($message);
	$smtp->dataend();
	$smtp->quit();
	return 1;
    }
    else
    {
	return 0;
    }
}

sub http_dump
{
#
# This sub takes one argument, the URL of the
# HTML to "dump".
#
# On success, returns the contents of the URL.
# Otherwise, it returns 0
#

    my $url = $_[0];
    my $ua = LWP::UserAgent->new();
    my $request = HTTP::Request->new(GET => "$url");
    my $response = $ua->request($request);
    if ($response->is_success)
    {
	return $response->content;
    }
    else
    {
	return 0;
    }
}

sub get_colors
{
#
# Takes no input
#
# Always returns 1
#
# set the following globals:
# $bgcolor - a hex triplet string for the background color
# $textcolor - a hex triplet string for the text color
# $table_bgcolor - a hex triplet for the background of the table
# $headimage - a string that contains tags necessary to put the
#              specified header image in the header
# $bgimage - a string that contains the necessary tag attributes
#            to set the background image in the body tag.
#
    if ($FORM{'bgcolor'})
    {
	my $tempcolor = $FORM{'bgcolor'};
	unless (length($tempcolor) == 6)
	{
	    $bgcolor = "#FFFFFF";
	}
	else
	{
	    $bgcolor = "#" . $tempcolor;
	}
    }
    else
    {
	$bgcolor = "#FFFFFF";
    }
    
    if ($FORM{'textcolor'})
    {
	my $tempcolor = $FORM{'textcolor'};
	unless (length($tempcolor) == 6)
	{
	    $textcolor = "#000000";
	}
	else
	{
	    $textcolor = "#" . $tempcolor;
	}
    }
    else
    {
	$textcolor = "#000000";
    }
    
    if ($FORM{'table_bgcolor'})
    {
	my $tempcolor = $FORM{'textcolor'};
	unless (length($tempcolor) == 6)
	{
	    $table_bgcolor = "";
	}
	else
	{
	    $table_bgcolor = "#" . $tempcolor;
	}
    }
    else
    {
	$table_bgcolor = $bgcolor;
    }
    
    if ($FORM{'head_image'})
    {
	$headimage = "<center><img src=$FORM{'head_image'} alt=Header></center>";
    }
    else
    {
	$headimage = "";
    }
    
    if ($FORM{'bg_image'})
    {
	$bgimage = "background=$FORM{'bg_image'}";
    }
    else
    {
	$bgimage = "";
    }
    return 1;
}

sub read_post
{
#
# Takes no arguments
#
# returns 1 upon success, 0 if
# read from stdin failed
#
# Creates global vars:
#
# %FORM - contains POST input keyed to
#         lower-case parameter names
#
# @itemnumbers - contains all item numbers
#         passed to the script by the POST
#         (i.e. (1, 2, 3) would result from 
#         having item1num, item2num, and 
#         item3num as POST values)
#
# @requiredfields - contains the names of
#         fields which are required, which
#         is checked by check_post()
#
# @statetaxes - contains the names of the
#         states for which tax should be
#         added.  Taxes are added by print_taxes
#
# @passthruvars - numbers corresponding to values
#                 to pass through as USERx if
#                 no_print_verified is true
#
# @passvarvars - numbers corresponding to variables
#                whose values are passed through as
#                USERx if no_print_verified is true
#
    my $buffer;
    if (read(STDIN, $buffer, $ENV{'CONTENT_LENGTH'}))
    {
	my @pairs = split(/&/, $buffer);
	foreach $pair (@pairs)
	{

# convert plus signs to spaces
	    $pair =~ s/\+/ /g;
	
# convert encoded data to characters	
	    $pair =~ s/%(..)/pack("c",hex($1))/ge;
	
# Finally, put the results in a hash, FORM
	    my $loc=index($pair,"=");
	    my $param=substr($pair, 0, $loc);
	    $param = "\L$param\E";   # Convert to all lowercase
	    my $value=substr($pair, $loc+1);
	    $FORM{$param} .= $value;
	    if ($param =~ /item(\d+)qty/)
	    {
		if ($FORM{$param}) { push @itemnumbers, $1; }
	    }
	    elsif ($param =~ /(.+)_required/)
	    {
		push @requiredfields, $1;
	    }
	    elsif ($param =~ /(.+)_tax/)
	    {
		push @statetaxes, $1;
	    }
	    elsif ($param =~ /passthru(\d)/)
	    {
		push @passthruvars, $1;
	    }
	    elsif ($param =~ /passvar(\d)/)
	    {
		push @passvarvars, $1;
	    }
	}
	return 1;
    }
    else
    {
	return 0;
    }
}

sub check_post
{
#
# Takes no arguments
#
# returns 1 if all required fields
# are filled in, 0 otherwise
#
    push @requiredfields, ("snd_name",
			   "snd_addr1",
			   "snd_city",
			   "snd_state",
			   "snd_zip",
			   "snd_country",
			   "snd_phone",
			   "snd_mail");
    
    foreach $field (@requiredfields)
    {
	if ($field)
	{
	    unless ($FORM{$field})
	    {
		return 0;
	    }
	}
    }
    return 1;
}

sub check_opts
{
#
# Takes no arguments
#
# returns 1 if all required
# admin options are present,
# 0 otherwise
#
    my @adminopts = ("result_url", 
		     "vt_url", 
		     "mail_company",
		     "mail_address", 
		     "mail_city",
		     "mail_state", 
		     "mail_zip",
		     "mail_recipient",
		     "script_url",
		     "local_country",
		     "mail_result_url",
		     "mailserver");
    
    foreach $option (@adminopts)
    {
	unless ($FORM{$option})
	{
	    return 0;
	}
    }
    unless ($FORM{'result_url_unverified'} || $FORM{'no_print_validated'})
    {
	return 0;
    }
    return 1;
}

sub check_price
{
#
# Takes no input
#
# Returns 1 if the price is 
# nonzero, 0 otherwise
#
# Sets the global
# variable $totalprice
# to the total price of
# the order
#
    $totalprice = 0;
    foreach $item (@itemnumbers)
    {
	my $price = $FORM{"item" . $item . "price"};
	my $qty = $FORM{"item" . $item . "qty"};
	$totalprice += $price * $qty;
    }
    unless ($totalprice)
    {
	return 0;
    }
    else
    {
	return 1;
    }
}

sub check_qty
{
#
# Takes no input
#
# Returns 1 if the total quantity
# is nonzero, 0 otherwise
#
# Sets the global variable $totalqty
# to the total quantity of all items
# in the order
#
    $totalqty = 0;
    foreach $item (@itemnumbers)
    {
	my $qtyname = "item" . $item . "qty";
	my $qty = $FORM{$qtyname};
	$totalqty += $qty;
    }
    unless ($totalqty)
    {
	return 0;
    }
    else
    {
	return 1;
    }
}

sub check_ship
{
#
# Takes no input
#
# Returns 1 if shipping price
# is nonzero, 0 otherwise
#
# Sets the global variable $totalship
# to the total shipping price for all
# items
#
    my $method;
    $methodtext = "";
    if ($FORM{'ship_method'})
    {
	$method = $FORM{'ship_method'};
    }
    else
    {
	$method = 1;
    }
    if ($FORM{"ship_method_" . $method . "_name"})
    {
	$methodtext = " (via " . $FORM{"ship_method_" . $method . "_name"} . ")";
    }
    $totalship = 0;
    if ($FORM{"global_ship$method"})
    {
	$totalship = determine_shipprice($FORM{"global_ship$method"}, $FORM{"global_iship$method"});
    }
    elsif ($FORM{"global_ship_each$method"})
    {
	my $shipprice = determine_shipprice($FORM{"global_ship_each$method"}, $FORM{"global_iship_each$method"});
        $totalship = $shipprice * $totalqty;
    }
    else
    {
	foreach $item (@itemnumbers)
	{
	    my $shipprice = determine_shipprice($FORM{"item" . $item . "ship$method"}, 
						$FORM{"item" . $item . "iship$method"});
	    $totalship += $FORM{"item" . $item . "qty"} * $shipprice;
	}
    }
    unless ($totalship)
    {
	return 0;
    }
    else
    {
	return 1;
    }
}

sub determine_shipprice
{
#
# Takes as arguments the variables for
# shipping and international shipping
#
# Returns the price of shipping depending
# on the country.
#
    my $country_local;
    if ($FORM{'snd_country'} eq $FORM{'local_country'})
    {
	$country_local = 1;
    }
    else
    {
	$country_local = 0;
    }
    unless ($_[1])
    {
	return $_[0];
    }
    else
    {
	if ($country_local)
	{
	    return $_[0];
	}
	else
	{
	    return $_[1];
	}
    }
}
    
sub print_taxes
{
#
# Takes no input
#
# Always returns 1
#

    $tax = 0;
STATE: foreach $state (@statetaxes)
    {
	if ($state eq "\L$FORM{'snd_state'}\E")
	{
	    $taxstate = "$state" . "_tax";
	    $taxinstate = $FORM{$taxstate};
	    $thetax = $taxinstate * $totalprice;
	    $thetax = int (0.5 + $thetax * 100) / 100;
	    $tax = $thetax;
	    $thetax = sprintf "%3.2f", $thetax;
	    chomp $thetax;
	    print "<tr><th colspan=4 align=right>Tax:</th>\n";
	    print "<th align=right><input type=hidden name=\"mail_tax\" value=\"$taxinstate\">\n";
	    print "\$$thetax </th></tr>\n";
	    last STATE;
	}
    }
    return 1;
}

sub print_items
{
#
# Takes 5 arguments:
# prefix - the prefix for items to be put as part of the <input> tag
# number_suffix - the suffix for the item_number field
# qty_suffix - the suffix for the item_qty field
# desc_suffix - the suffix for the item_desc field
# price_suffix - the suffix for the item_price field
#
# Always returns 1
#
    my $item_count = 1;
    print <<END;
<table border=0 bgcolor=$table_bgcolor align=center>
<tr>
<th>Quantity</th>
<th>Item #</th>
<th>Description</th>
<th>Price Each</th>
<th>Item Total</th>
</tr>
END
    foreach $item (@itemnumbers)
    {
	my $itemdesc = $FORM{"item" . $item . "desc"};
	my $itemprice = $FORM{"item" . $item . "price"};
	my $itemnum = $FORM{"item" . $item . "num"};
	my $itemqty = $FORM{"item" . $item . "qty"};
	my $itemtotalprice = $itemprice * $itemqty;
	
	$itemprice = sprintf "%3.2f", $itemprice;
	chomp $itemprice;

	$itemtotalprice = sprintf "%3.2f", $itemtotalprice;
	chomp $itemtotalprice;
	
	my $form_item_number = $_[0] . $item_count . $_[1];
	my $form_item_qty = $_[0] . $item_count . $_[2];
	my $form_item_desc = $_[0] . $item_count . $_[3];
	my $form_item_price = $_[0] . $item_count . $_[4];
	$item_count++;
	print <<END;
<tr>
<td align=center>
$itemqty
<input type="hidden" name="$form_item_qty" value="$itemqty">
</td>
<td align=center>
$itemnum
<input type="hidden" name="$form_item_number" value="$itemnum">
</td>
<td align=center>
$itemdesc
<input type="hidden" name="$form_item_desc" value="$itemdesc">
</td>
<td align=right>
\$$itemprice
<input type="hidden" name="$form_item_price" value="$itemprice">
</td>
<th align=right>
\$$itemtotalprice
</th>
</tr>
END
    }
    return 1;
}

sub print_bill
{
    print <<END;
<tr>
<th colspan=2 align=right>Name:</th>
<td colspan=3 align=left><input type="text" name="snd_name" value="$FORM{'snd_name'}" size=40></td>
</tr>
<tr>
<th colspan=2 align=right>Address:</th>
<td colspan=3 align=left><input type="text" name="snd_addr1" value="$FORM{'snd_addr1'}" size=40></td>
</tr>
<tr>
<th colspan=2 align=right>City:</th>
<td colspan=3 align=left><input type="text" name="snd_city" value="$FORM{'snd_city'}" size=40></td>
</tr>
<tr>
<th colspan=2 align=right>State:</th>
<td colspan=3 align=left><input type="text" name="snd_state" value="$FORM{'snd_state'}" size=8 maxlength=2>
    &nbsp;&nbsp;&nbsp;&nbsp; <strong>Zip:</strong> 
	<input type="text" name="snd_zip" value="$FORM{'snd_zip'}" size=20></td>
</tr>
<tr>
<th colspan=2 align=right>Country:</th>
<td colspan=3 align=left><input type="text" name="snd_country" value="$FORM{'snd_country'}" size=40></td>
</tr>
<tr>
<th colspan=2 align=right>Phone:</th>
<td colspan=3 align=left><input type="text" name="snd_phone" value="$FORM{'snd_phone'}" size=40></td>
</tr>
<tr>
<th colspan=2 align=right>Email:</th>
<td colspan=3 align=left><input type="text" name="snd_mail" value="$FORM{'snd_mail'}" size=40></td>
</tr>
END
    return 1;
}

sub print_ship
{
    print <<END;
<tr><th colspan=5 align=left>Please enter your shipping address:</th></tr>
<tr>
<th colspan=2 align=right>Name:</th>
<td colspan=3 align=left><input type="text" name="shp_name" size=40></td>
</tr>
<tr>
<th colspan=2 align=right>Address:</th>
<td colspan=3 align=left><input type="text" name="shp_addr1" size=40></td>
</tr>
<tr>
<th colspan=2 align=right>City:</th>
<td colspan=3 align=left><input type="text" name="shp_city" size=40></td>
</tr>
<tr>
<th colspan=2 align=right>State:</th>
<td colspan=3 align=left><input type="text" name="shp_state" size=8 maxlength=2>
    &nbsp;&nbsp;&nbsp;&nbsp; <strong>Zip:</strong> 
	<input type="text" name="shp_zip" size=20></td>
</tr>
<tr>
<th colspan=2 align=right>Country:</th>
<td colspan=3 align=left><input type="text" name="shp_country" size=40></td>
</tr>
<tr>
<th colspan=2 align=right>Phone:</th>
<td colspan=3 align=left><input type="text" name="shp_phone" size=40></td>
</tr>
END
    return 1;
}

sub print_header
{
#
# Takes one argument, the header text
# 
# Always returns 1
#
    print <<END;
Content-type: text/html

<html><head><title>$_[0]</title></head>
<body bgcolor=$bgcolor text=$textcolor $bgimage>
$headimage
END
    return 1;
}

sub print_form_setup
{
#
# Takes no arguments
#
# Always returns 1
#
    my $returntourl;
    my @userxvars;
    my $headimage_escaped;
    my $bgimage_escaped;
    if ($FORM{'no_print_validated'})
    {
	$returntourl = $FORM{'result_url'};
	foreach $var (@passthruvars)
	{
	    my $varval = $FORM{'passthru' . $var};
	    my $varname = "USER" . $var;
	    push @userxvars, "<input type=hidden name=\"$varname\" value=\"$varval\">";
	}
	foreach $var (@passvarvars)
	{
	    my $varval = $FORM{"\L" . $FORM{"passvar" . $var} . "\E"};
	    my $varname = "USER" . $var;
	    push @userxvars, "<input type=hidden name=\"$varname\" value=\"$varval\">";
	}
    }
    else
    {
	$returntourl = $FORM{'script_url'};
	$headimage_escaped = $headimage;
	$bgimage_escaped = $bgimage;
	$bgimage_escaped =~ s/=/-:-/g;
	$headimage_escaped =~ s/</-lt-/g;
	$headimage_escaped =~ s/>/-gt-/g;
	$headimage_escaped =~ s/ /^/g;
	@userxvars = ("<input type=hidden name=USER1 value=\"$bgcolor\">",
		      "<input type=hidden name=USER2 value=\"$textcolor\">",
		      "<input type=hidden name=USER3 value=\"$headimage_escaped\">",
		      "<input type=hidden name=USER4 value=\"$bgimage_escaped\">",
		      "<input type=hidden name=USER5 value=\"$FORM{result_url}\">",
		      "<input type=hidden name=USER6 value=\"$FORM{result_url_unverified}\">");
    }
    print <<END;
<form method=POST action="$FORM{'vt_url'}">
<input type=hidden name=ReturnTo value="$returntourl">
<input type=hidden name=mail_company value="$FORM{'mail_company'}">
<input type=hidden name=mail_address value="$FORM{'mail_address'}">
<input type=hidden name=mail_city value="$FORM{'mail_city'}">
<input type=hidden name=mail_state value="$FORM{'mail_state'}">
<input type=hidden name=mail_zip value="$FORM{'mail_zip'}">
<input type=hidden name=mail_recipient value="$FORM{'mail_recipient'}">
END
    foreach $userx (@userxvars)
    {
	print "$userx \n";
    }
}
	    
sub returning_from_mailto
{
    my $message = "$FORM{'mail_company'} Order \n\n";
    foreach $item (@itemnumbers)
    {
	my $itemnumber = $FORM{"item" . $item . "num"};
	my $itemdesc = $FORM{"item" . $item . "desc"};
	my $itemqty = $FORM{"item" . $item . "qty"};
	$message .= "Item:  $itemnumber \n";
	$message .= "Description:  $itemdesc \n";
	$message .= "Quantity:  $itemqty \n";
	$message .= "\n --------- \n\n";
    }
    if (send_mail($message, 
		  "$FORM{'mail_company'} Order", 
		  $FORM{'snd_mail'}, 
		  $FORM{'mail_recipient'}, 
		  $FORM{'cc_recipient'}))
    {
	print <<END;
HTTP 1.1/200 OK
Location: $FORM{"mail_result_url"}
Content-type: text/html

END
    }
    else
    {
	print <<END;
Content-type:text/html

<html><head><title>Mailer Error</title></head>
<body>
<h1>Mailer Error</h1>
<p>Unable to send mail message.  Please e-mail the <a href="mailto:$FORM{'mail_recipient'}">site contact</a>
concerning the problem.
</body></html>
END
    }
    exit;    
}

sub returning_from_TMM
{
#
# Takes no input
#
# Kills the script when finished
#
    $bgcolor = $FORM{'user1'};
    $textcolor = $FORM{'user2'};
    $headimage = $FORM{'user3'};
    $bgimage = $FORM{'user4'};
    $bgimage =~ s/-:-/=/g;
    my $result_url = $FORM{'user5'};
    my $result_url_unvalidated = $FORM{'user6'};
    $headimage =~ s/-lt-/</g;
    $headimage =~ s/-gt-/>/g;
    $headimage =~ s/\^/ /g;
    if ($FORM{'validated'} eq "Y")
    {
	$other_html = http_dump($result_url);
	print_header("Credit Card Validation Successful");
	print <<END;
<p>Your credit card was successfully validated.  Your approval code is $FORM{'app_code'}.
$other_html
<hr width=50% size=1>
<p>
<font size=1 color=green>
This page was generated by TMMOrder v.$version<br>
Written by Riad Wahby &lt;<a href="mailto:rsw\@phonon.net">rsw\@phonon.net</a>&gt;<br>
Available at <a href="ftp://$ftpservername/pub/tmmorder">ftp://$ftpservername/pub/tmmorder</a>
</font></center></body></html>
END
    }
    else
    {
	$other_html = http_dump($result_url_unvalidated);
	print_header("Error: Credit Card Not Validated");
	print <<END;
<p>Your credit card was not validated.
$other_html
<hr width=50% size=1>
<p><center>
<font size=1 color=green>
This page was generated by TMMOrder v.$version<br>
Written by Riad Wahby &lt;<a href="mailto:rsw\@phonon.net">rsw\@phonon.net</a>&gt;<br>
Available at <a href="ftp://$ftpservername/pub/tmmorder">ftp://$ftpservername/pub/tmmorder</a>
</font></center></body></html>
END
    }
    exit;
}

sub mail_order_dispatch
{
#
# Takes no arguments
#
# This sub kills the script when it is
# finished running
#
    print_header("Please verify your selections");
    print <<END;
<p>Please verify this information.  If there are any mistakes in the items,
please use the back button in your browser to fix them.  You can make corrections
to your mailing address here.
<form method=POST action=$FORM{'script_url'}>
<input type="hidden" name="mailto" value="true">
<input type="hidden" name="mail_result_url" value="$FORM{'mail_result_url'}">
<input type="hidden" name="mail_company" value="$FORM{'mail_company'}">
<input type="hidden" name="mail_recipient" value="$FORM{'mail_recipient'}">
<input type="hidden" name="cc_recipient" value="$FORM{'cc_recipient'}">
<input type="hidden" name="mailserver" value="$FORM{'mailserver'}">
<table border=0 align=center>
END
    print_items("item", "num", "qty", "desc", "price");
    print "<tr><th colspan=5 align=left>This order will ship to:</th></tr>\n";
    print_bill();
    print <<END;
<tr><th height=50 valign=middle colspan=5><input type="submit" value="This information is correct"></th></tr>
</table></form>
<hr width=50% size=1>
<p><center>
<font size=1 color=green>
This page was generated by TMMOrder v.$version<br>
Written by Riad Wahby &lt;<a href="mailto:rsw\@phonon.net">rsw\@phonon.net</a>&gt;<br>
Available at <a href="ftp://$ftpservername/pub/tmmorder">ftp://$ftpservername/pub/tmmorder</a>
</font></center>
</body></html>
END
    exit;
}

sub TMM_order_dispatch
{
#
# Takes no input
#
# Always kills the program when it finishes
#
    print_header("Please verify your selections");
    print <<END;
<p>Please verify your selections.  If there are any mistakes in the items, please use
the back button in your browser to fix them.  You can make any changes and corrections
to your mailing address here.
END
    print_form_setup();
    print_items("A", "N", "Q", "P", "");
    print_taxes();
    if (check_ship())
    {
	$shipprint = sprintf "%3.2f", $totalship;
	chomp $shipprint;
	print <<END;
<tr>
<th colspan=4 align=right>Shipping$methodtext:</th>
<th align=right><input type=hidden name=ShipTotal value="$shipprint">\$$shipprint</th></tr>
END
    }
    $totalprint = $totalprice + $totalship + $tax;
    $totalprint = sprintf "%3.2f", $totalprint;
    chomp $totalprint;
    print <<END;
<tr>
<th colspan=4 align=right>Total Price:</th>
<th align=right><input type=hidden name=snd_total value="$totalprint">\$$totalprint</th></tr>
END
    if ($FORM{'diffship'})
    {
	print "<tr><th colspan=5 align=left>Your order will be billed to:</th></tr>\n";
	print_bill();
	print_ship();
    }
    else
    {
	print "<tr><th colspan=5 align=left>The address for billing and shipping is:</th></tr>\n";
	print_bill();
    }
    print <<END;
<tr><th colspan=5 height=50 valign=middle><input type=submit value="This information is correct."></th></tr>
</table>
</form>
<hr width=50% size=1>
<p><center>
<font size=1 color=green>
This page was generated by TMMOrder v.$version<br>
Written by Riad Wahby &lt;<a href="mailto:rsw\@phonon.net">rsw\@phonon.net</a>&gt;<br>
Available at <a href="ftp://$ftpservername/pub/tmmorder">ftp://$ftpservername/pub/tmmorder</a>
</font></center>
</body>
</html>
END
    exit;
}
